--
-- OrganizedPanic
-- 
-- Copyright (c) 2021 Dana Basken
--

local VERSION="9.1.0.1"

local _, NS = ...

OrganizedPanic = LibStub("AceAddon-3.0"):NewAddon(NS.ADDON_NAME, "AceConsole-3.0", "AceTimer-3.0", "AceEvent-3.0", "AceComm-3.0")

print(PrefixString(NS.ADDON_NAME, ("v%s"):format(VERSION), NS.PREFIX_COLOR))

local OPTIONS = NS:BuildOptions(NS.ADDON_NAME, OrganizedPanic)

local AceGUI = LibStub("AceGUI-3.0") -- needed for AceConfigDialog
local AceEvent = LibStub("AceEvent-3.0")
local AceTimer = LibStub("AceTimer-3.0")

function OrganizedPanic:OnInitialize()
  self.db = LibStub("AceDB-3.0"):New("OrganizedPanicDB", NS.DEFAULTS, true)

  LibStub("AceConfig-3.0"):RegisterOptionsTable(NS.ADDON_NAME, OPTIONS)
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions(NS.ADDON_NAME)
  
  self:CreateFlasher()

  self.castingStates = {}
  self.name = UnitName("player")

  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", "SPELLCAST_EVENT_HANDLER")
  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_START", "SPELLCAST_EVENT_HANDLER")
  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_STOP", "SPELLCAST_EVENT_HANDLER")
  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START", "SPELLCAST_EVENT_HANDLER")
  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP", "SPELLCAST_EVENT_HANDLER")

  OrganizedPanic:RegisterChatCommand(NS.CHAT_COMMAND_NAME, "HandleChatCommand")
end

function OrganizedPanic:OnEnable()
end

function OrganizedPanic:SPELLCAST_EVENT_HANDLER(event, unit, guid, id)
  local caster = UnitName(unit)
  local originalEvent = event
  if event == "UNIT_SPELLCAST_SUCCEEDED" or event == "UNIT_SPELLCAST_START" or event == "UNIT_SPELLCAST_CHANNEL_START" then
    event = "SPELL_START"
  end
  if event == "UNIT_SPELLCAST_STOP" or event == "UNIT_SPELLCAST_CHANNEL_STOP" then
    event = "SPELL_STOP"
  end
  if self.db.profile.debug then
    print(("%s (%s): %s %s %s"):format(event, originalEvent, caster, id, GetSpellInfo(id)))
  end
  if self:SpellCanPanic(caster, id) then
    self:SetCastingState(caster, event, id)
  end
end

function OrganizedPanic:SpellCanPanic(caster, id)
  -- does the spell have a "panic" entry?
  if not NS.PANIC_SPELLS[id] then return end
  -- if including self, act like we're in raid and bypass "self" check
  if self.db.profile.includeSelf then return true end
  -- check for raid group
  if not self:InRaid() then return end
  -- check to see if caster is me
  return caster ~= self.name
end

function OrganizedPanic:InRaid()
  local isInInstance, instanceType = IsInInstance();
  return instanceType == "raid"
end

function OrganizedPanic:SetCastingState(caster, event, id)
  local now = GetTime()
  if not self.castingStates[caster] then self.castingStates[caster] = {event = nil, spell = nil, when = now} end
  local lastState = self.castingStates[caster]
  if lastState.spell ~= id or lastState.event ~= event or now - lastState.when > 30 then
    self.castingStates[caster] = {event = event, spell = id, when = now}
    self:OnCastingStateChanged(caster)
  end
end

function OrganizedPanic:OnCastingStateChanged(caster)
  local state = self.castingStates[caster]
  if self.db.profile.debug then
    print(("STATE CHANGE: %s %s %s %d"):format(caster, state.event, state.spell, state.when))
  end
  if state.event == "SPELL_START" then
    if self.db.profile.includeRole then
      if UnitGroupRolesAssigned("player") == UnitGroupRolesAssigned(caster) then
        self:Panic(caster, state.spell)
      end
    else
      self:Panic(caster, state.spell)
    end
  end
end

function OrganizedPanic:Panic(caster, id)
  self:PlayAlertSound()
  self:ShowRaidWarning(caster, id)
  self:FlashBorder()
end

function OrganizedPanic:CreateFlasher()
  self.frame = CreateFrame("Frame", nil, UIParent, BackdropTemplateMixin and "BackdropTemplate")
  self.frame:SetFrameStrata("BACKGROUND");
  self.frame:SetClampedToScreen(true)
  self.frame:SetAllPoints(UIParent)

  local texture = self.frame:CreateTexture(nil,"BACKGROUND");
  texture:SetTexture("Interface\\Addons\\OrganizedPanic\\images\\border.tga");
  texture:SetBlendMode("ADD");
  texture:SetVertexColor(self.db.profile.borderColor.r, self.db.profile.borderColor.g, self.db.profile.borderColor.b);
  texture:SetAllPoints(self.frame);
  self.frame.texture = texture;

  self.frame:SetPoint("CENTER",0,0);
  self.frame:Hide();

  self.animationGroup = self.frame:CreateAnimationGroup()
  self.animationGroup:SetLooping("BOUNCE")

  local animation = self.animationGroup:CreateAnimation("Alpha")
  animation:SetDuration(0.25)
  animation:SetFromAlpha(0.1)
  animation:SetToAlpha(0.9)
end

function OrganizedPanic:PlayAlertSound()
  if self.db.profile.playSound then
    PlaySoundFile(NS.SOUNDS[self.db.profile.sound], "Master")
  end
end

function OrganizedPanic:ShowRaidWarning(caster, id)
  local spell = GetSpellInfo(id)
  if not spell then spell = id end
  local message = ("%s started casting %s"):format(caster, spell)
  RaidNotice_AddMessage(RaidWarningFrame, message, ChatTypeInfo["RAID_WARNING"])
end

function OrganizedPanic:FlashBorder()
  if self.db.profile.flashBorder then
    self:StartFlashing()
  end
end

function OrganizedPanic:StartFlashing()
  if self.db.profile.debug then print("StartFlashing") end
  if self.animationGroup:IsPlaying() then self:StopFlashing() end
  self.frame.texture:SetVertexColor(self.db.profile.borderColor.r, self.db.profile.borderColor.g, self.db.profile.borderColor.b);
  self.animationGroup:Play()
  self.frame:Show()
  if self.flashTicker then self.flashTicker:Cancel() end
  self.flashTicker = C_Timer.NewTicker(1, function() self:StopFlashing() end, 0)
end

function OrganizedPanic:StopFlashing()
  if self.db.profile.debug then print("StopFlashing") end
  if self.flashTicker then self.flashTicker:Cancel() end
  self.flashTicker = nil
  self.frame:Hide()
  self.animationGroup:Stop()
end

function OrganizedPanic:GetOption(info)
  -- print(ObjectToString(info.option))
  local property = info.option.arg
  return self.db.profile[property]
end

function OrganizedPanic:SetOption(info, value)
  -- print(ObjectToString(info.option), value)
  local property = info.option.arg
  self.db.profile[property] = value
end

function OrganizedPanic:SetBorderColor(info, r, g, b, a)
  self.db.profile.borderColor.r = r
  self.db.profile.borderColor.g = g
  self.db.profile.borderColor.b = b
  self:StartFlashing()
end

function OrganizedPanic:GetBorderColor(info)
  return self.db.profile.borderColor.r, self.db.profile.borderColor.g, self.db.profile.borderColor.b, 1.0
end

function OrganizedPanic:AddonPrint(format, ...)
  print(PrefixString(NS.ADDON_NAME, string.format(format, ...), NS.PREFIX_COLOR))
end

function OrganizedPanic:HandleChatCommand(input)
  if not input or input:trim() == "" then
    LibStub("AceConfigDialog-3.0"):Open("OrganizedPanic")
  else
    if input == "help" then
      self:AddonPrint("%s", "Commands:")
    elseif input == "options" then
      print(ObjectToString(self.db.profile))
    elseif input == "panic" then
      self:Panic(self.name, "Debug")
    elseif input == "namespace" then
      print(ObjectToString(NS))
    else
      self:AddonPrint("unknown command: '%s', try '/%s help'", input, NS.CHAT_COMMAND_NAME)
    end
  end
end
